#! /usr/bin/perl -T

use 5.010;
use strict;
use warnings;

use Data::Dumper;

use CGI qw(:all);
use CGI::Carp qw(fatalsToBrowser);
use Crypt::Simple;
use Digest::SHA1 qw(sha1_hex);
use Mail::Sendmail;
use DBI;
use Template;
use File::Basename;
use Mail::RFC822::Address qw(valid);

sub insert(\%);
sub confirm($$);
sub read_conf($);

sub do_invite();
sub page;
sub mail;

sub _encrypt {
    my $_ = encrypt(shift);
    s/\+/-/g;
    s/\//_/g;
    return $_;
}

sub _decrypt {
    my $_ = shift;
    s/_/\//g;
    s/-/+/g;
    return decrypt($_);
}


delete @ENV{grep /PATH$/ => keys %ENV};
$ENV{PATH} = "/usr/bin:/usr/sbin:/bin:/sbin";

my $DSN = "DBI:SQLite:var/db.sqlite3";
my $EXPIRATION = 60;  # expiration time of the first link
my $SECRET = read_conf "conf/secret";
my $SELF = read_conf "conf/self";   
my $SENDER = read_conf "conf/sender";
my $SENDMAIL = "/usr/sbin/sendmail -f $SENDER -t";

my %FIELDS = (
    MAN => [qw[givenname surname email]],
    OPT => [qw[tel]]
);


my %ttconfig = (
    INCLUDE_PATH => "templates",
    VARIABLES => {
	SELF => $SELF,
    },
);

my $DBH = DBI->connect($DSN, undef, undef, {RaiseError=>1});
END { $DBH and $DBH->disconnect }

MAIN: {

    # Redirect if called without the script name, this avoids
    # other problems later
    if (basename($ENV{SCRIPT_NAME}) ne basename(url(-path_info => 0, -relative => 1))) {
	print redirect(basename($ENV{SCRIPT_NAME}));
    }

    # ACCESS
    if (!path_info()) {
        do_invite();
	exit 0;
    }

    # /done.<uuid>/ filled forms etc
    # /done/
    if (path_info() =~ /^\/?done(?:\.(?<uuid>.*?))?\/(?<step>.*)$/) {
	my %x = %+;
	if ($x{step} ~~ [qw(invitation form confirmation)]) {	
	    eval {
		page("html.$x{step}.done.tpl", {
		    done => {
			map({ ($_, _decrypt(param($_))) } param()),
		    },
		    uuid => UUID->new($x{uuid}, map { @$_ } values %FIELDS),
		});
		exit 0;
	    };
	    die $@ if $@;
	    if ($@) {
		page("html.denied.tpl");
		exit 0;
	    }
	}
    }

    # /show/(…)
    if (path_info() =~ /^\/?show\/(.*)$/) {
	if ($1 ~~ [qw(info)]) {
	    page("$1.tpl");
	    exit 0;
	}
	page("html.denied.tpl");
	exit 0;
    }

    # /user.<uuid>
    if (path_info() =~ /^\/?user\.(.*)(?:\/(.*))?$/) {
	my $uuid = $1;

	if (param("confirm")) {

	my $confirmed = param("confirm") eq "yes" ? 1 : 0;   
	my %data = confirm($uuid => $confirmed);

	if ($data{error}) {
	    page("html.denied.tpl");
	    exit 0;
	}

	mail("mail.confirmed.tpl", {
	    to => $data{email},
	    confirmed => $confirmed,
	});

	print redirect("$SELF/done.$uuid/confirmation?"
	    . "confirmed=" . _encrypt($confirmed));

	exit 0;
	}

    }

    # /tmp.<uuid>
    if (path_info() =~ /^\/?tmp\.(.*)$/) {
	my $_ = $1;

	eval {
	    my $time = _decrypt($_) or die "DECRYPTION ERROR";
	    $time =~ /^\d+$/ or die "FORMAT ERROR";
	    time() - $time < (60 * $EXPIRATION) or die "EXPIRED";
	}; if ($@) {
	    page("html.denied.tpl");
	    exit 0;
	}

	my %warn;
	my %value;

	if (param("clear")) {
	    Delete_all();
	}

	# Submission
	if (param("submit")) {
	    foreach my $param (@{$FIELDS{MAN}}, @{$FIELDS{OPT}}) {

		# strip away spaces to "untaint" the variables, additionally
		# limit the length
		my $_ = param($param);
		/^\s*(.*)\s*$/;

		if (!length $1 and $param ~~ @{$FIELDS{MAN}}) {
		    push @{$warn{$param}}, "Leer!?";
		}

		if (length $1 > 200) {
		    push @{$warn{$param}}, "Zu lang (bitte weniger als 200 Zeichen).";
		}

		param(-name => $param, value => $1);
		$value{$param} = $1;
	    }

	    # Mail needs extra check
	    if ($value{email} and not valid($value{email})) {
		push @{$warn{email}}, "Keine gültig erscheinende Mail-Adresse.";
	    }

	    foreach (keys %warn) {
		$warn{$_} = join " ", @{$warn{$_}};
	    }

	    if (%warn) {
		page("html.form.tpl", {
		    warn => %warn ? \%warn : undef,
		    value => {
			givenname => scalar param("givenname"),
			surname => scalar param("surname"),
			email => scalar param("email"),
		    }
		});
		exit 0;
	    }

	    my %r = insert(%value);

	    mail("mail.form.done.tpl", {
		to => $value{email},
		url => {
		    yes => "$SELF/user.$r{uuid}?confirm=yes",
		    no  => "$SELF/user.$r{uuid}?confirm=no",
		}
	    });

	    print redirect("$SELF/done.$r{uuid}/form?"
		. "exists=" . _encrypt($r{exists}//0));
	    exit 0;
	}

	page("html.form.tpl");
	exit 0;
    }

    page("html.denied.tpl", {
	    url => $SELF,
    });
    exit 0;
}

sub insert(\%) {
    my %value = %{$_[0]};
    my $uuid = sha1_hex($SECRET . 
	join "\0" => @value{@{$FIELDS{MAN}}});

    $DBH->begin_work;
	my $sth;
	$sth = $DBH->prepare("SELECT created FROM db WHERE uuid = ?");
	$sth->execute($uuid);

	if (my $r = $sth->fetchrow_hashref) {
	    my $created = $r->{created};
	    $DBH->rollback;
	    return (uuid => $uuid,
	            exists => $created);
	}

	local $" = ", ";
	$sth = $DBH->prepare("INSERT INTO db (@{$FIELDS{MAN}}, @{$FIELDS{OPT}}, uuid, created)
		VALUES(?, ?, ?, ?, ?, datetime('now'))");
	$sth->execute(@value{@{$FIELDS{MAN}}, @{$FIELDS{OPT}}}, $uuid);
    $DBH->commit;

    return (uuid => $uuid);
}

sub confirm($$) {
    my ($uuid, $confirmed) = @_;
    my %data;

    $DBH->begin_work;
	
	local $" = ", ";
	my $sth = $DBH->prepare("SELECT @{$FIELDS{MAN}}, @{$FIELDS{OPT}}, confirmed FROM db WHERE uuid = ?");
	$sth->execute($uuid);
	my $r = $sth->fetchrow_hashref;
	if (!$r) {
	    $DBH->rollback;
	    return (error => "NOT FOUND");
	}
	@data{@{$FIELDS{MAN}}} = @{$r}{@{$FIELDS{MAN}}};
	@data{@{$FIELDS{OPT}}} = @{$r}{@{$FIELDS{OPT}}};


	if ($confirmed) {
	    $sth = $DBH->prepare("UPDATE db SET confirmed = datetime('now')  WHERE uuid = ?");
	}
	else {
	    $sth = $DBH->prepare("DELETE FROM db WHERE uuid = ?");
	}
	$sth->execute($uuid);
	
    $DBH->commit;

    return %data;
}

sub read_conf($) {
    local $/ = undef;
    open(my $f, $_[0]) or die "Can't open $_[0]: $!\n";
    my $_ = <$f>;
    s/^#.*//mg;
    /^\s*(.*?)\s*$/s;
    return $1;
}

sub do_invite() {
	my %warn;
	my $sent;

	if (param("email")) {
	    if (not valid param("email")) {
		$warn{email} = "INVALID";
	    }
	    else {
		my $xxx = _encrypt(time);

		mail("mail.invitation.tpl", {
		    to   => scalar(param("email")),
		    url  => "$SELF/tmp.$xxx"});

		$sent = param("email");
	    }
	    print redirect("$SELF/done/invitation"
			. "?email=" . _encrypt($sent)
			. ";expires=" . _encrypt($EXPIRATION));
	    exit 0;
	}
	
	page("html.invitation.tpl", {
	    sent => $sent,
	    warn => %warn ? \%warn : undef,
	    expires => $EXPIRATION,
	    value => { email => scalar param("email") },
	});
}


sub page {
    state $tt = Template->new(\%ttconfig);
    print header(-charset => "UTF-8");
    $tt->process(@_);
}

sub mail {
    state $tt = Template->new(\%ttconfig);
    open(my $sendmail, "|$SENDMAIL") 
	or die "Can't open $SENDMAIL: $!\n";
    $tt->process(@_, $sendmail)
    or die $tt->error();
    close($sendmail) 
	or die "SENDMAIL: $!\n";
}

{ package UUID;
  use strict;
  use warnings;

  sub new {
    my $self = bless {} => shift;
    $self->{uuid} = shift;
    $self->{fields} = [@_];
    my $sth = $DBH->prepare("SELECT "
	. join(", " => @{$self->{fields}})
	. " FROM db WHERE uuid = ?");
    $sth->execute($self->{uuid});
    $self->{r} = $sth->fetchrow_hashref;
    $sth->finish;
    return undef if not $self->{r};
    return $self;
  }

  sub AUTOLOAD {
    my $self = shift;
    my ($f) = ($UUID::AUTOLOAD =~ /.*::(.*)/);
    return $self->{r}{$f};
  }
}

